#define _WIN32_WINNT   0x0400
#include <windows.h>
#include <map>

using namespace std;

#include "FileAPIOverride.h"
#include "ThreadAPIOverride.h"
#include "ErrorDescriber.h"
#include "debugio.h"
#include "Locker.h"
#include "wcconv.h"
#include "Log.h"

typedef struct {
	SIZE_T stack_size;
	LPTHREAD_START_ROUTINE function;
	LPVOID arg;
	DWORD flags;
	DWORD id;
} ThreadData;

typedef map<HANDLE, ThreadData> ThreadList;

static class _thd : private Locker {
	public:
		void NewThread(HANDLE h, const ThreadData &td) {
			Lock();
			thds[h] = td;
			Unlock();
		}

		bool RemoveThread(HANDLE h) {
			bool rv = false;
			Lock();
			ThreadList::iterator i = thds.find(h);
			if (i != thds.end()) {
				thds.erase(i);
				rv = true;
			}
			Unlock();
			return rv;
		}

		bool Exists(HANDLE h) {
			bool e;
			Lock();
			e = thds.find(h) != thds.end();
			Unlock();
			return e;
		}

		ThreadData Get(HANDLE h) {
			ThreadData tmp = {0};

			Lock();
			ThreadList::iterator i = thds.find(h);
			if (i != thds.end())
				tmp = i->second;
			Unlock();

			return tmp;
		}

		void AddDefaults(Report &r) {
			Lock();
			r.AddInt("num_threads", (int)thds.size());
			r.AddInt("current_thread", (int)GetCurrentThreadId());
			Unlock();
		}

	private:
		ThreadList thds;
} thd;

HANDLE __stdcall nAPI_CreateThread(LPSECURITY_ATTRIBUTES lpTA, SIZE_T stacksz, LPTHREAD_START_ROUTINE f, LPVOID arg, DWORD flags, LPDWORD threadid) {
	Report r("CreateThread");
	HANDLE retval;
	DWORD tid = 0;

	retval = CreateThread(lpTA, stacksz, f, arg, flags, threadid);
	if (retval) {
		tid = threadid ? *threadid : -1;

		const ThreadData tdtoa = {stacksz, f, arg, flags, tid};
		thd.NewThread(retval, tdtoa);
	} else {
		r.AddString("greska", ErrorDescription());
	}
	dprintf("CreateThread(%p, %u, %p, %p, 0x%08X, %d) = %p\n", lpTA, (unsigned int)stacksz, f, arg, (unsigned int)flags, (unsigned int)tid, retval);

	r.AddInt("stack", (int)stacksz);
	r.AddPtr("function", (void*)f);
	r.AddHex("flags", flags);
	r.AddInt("ID", tid);
	thd.AddDefaults(r);
	AddToLog(r);

	return retval;
}

VOID __stdcall nAPI_ExitThread(DWORD exit_code) {
	Report r("ExitThread");
	
	r.AddInt("exit_code", exit_code);
	thd.AddDefaults(r);
	AddToLog(r);

	ExitThread(exit_code);
}

BOOL __stdcall nAPI_CloseHandle(HANDLE object) {
	if (!thd.Exists(object)) {
		dprintf("%p nije thread, mozda je fajl?\n", object);
		return nAPI_CloseHandleFile(object);
	}

	Report r("CloseHandle");
	BOOL retval;
	bool uspjeh = false;
	BOOL imaexit;
	DWORD ec = 0;
	
	imaexit = GetExitCodeThread(object, &ec);

	retval = CloseHandle(object);
	dprintf("CloseHandle(%p) = %d\n", object, (int)retval);

	const ThreadData td = thd.Get(object);

	if (retval) {
		uspjeh = thd.RemoveThread(object);
	} else {
		r.AddString("greska", ErrorDescription());
	}

	if (imaexit) r.AddInt("exit_code", (int)ec);
	r.AddString("type", "THREAD_HANDLE");
	r.AddInt("ID", (int)td.id);
	r.AddInt("ok", uspjeh);
	r.AddPtr("handle", object);
	thd.AddDefaults(r);
	AddToLog(r);

	return retval;
}

// ne previse vezano uz threadove...

static void nAPI_GetModuleHandle(const char *lpModName, HMODULE retval) {
	Report r("GetModuleHandle");

	dprintf("GetModuleHandle(%s) = %p\n", lpModName, retval);

	if (!retval) {
		r.AddString("greska", ErrorDescription());
	}

	r.AddString("name", lpModName);
	r.AddPtr("handle", retval);

	AddToLog(r);
}

HMODULE __stdcall nAPI_GetModuleHandleA(LPCSTR lpModName) {
	HMODULE retval = GetModuleHandleA(lpModName);
	nAPI_GetModuleHandle(lpModName, retval);
	return retval;
}

HMODULE __stdcall nAPI_GetModuleHandleW(LPCWSTR lpModName) {
	HMODULE retval = GetModuleHandleW(lpModName);
	nAPI_GetModuleHandle(wcconvs(lpModName).c_str(), retval);
	return retval;
}
